#####################################################################
#:ClearQuest Web Java interface object
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2008. All Rights Reserved.
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

=head1 NAME

CQCC::CQWebJava - Subclass of CQCC::CQuest used to manage communications
with ClearQuest using the CQWeb Java interface.

=head1 CLASS DESCRIPTION

This class implements the necessary methods to logon to ClearQuest,
perform queries, and modify records over the CQWeb Java interface. 

It no longer has any dependency on the older CQWeb ASP support class.

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

######################################################################
#::Package CQCC::CQWebJava
######################################################################

package CQCC::CQWebJava;
use CQCC::CQuest;
@ISA = qw( CQCC::CQuest );
use strict;
sub Version { return '2.2' };

use Socket;
use Sys::Hostname;
use URI::Escape;
use CQCC::IOBase;
use CQCC::Parm;
use CQCC::Timer;
use CQCC::CQSchema;
use constant IPV6_IS_SUPPORTED => 
	eval { require Socket6 } &&
	eval { require IO::Socket::INET6 };

require Config;

######################################################################
#::Class Globals 
######################################################################

my $className = "CQWebJava"; 
my $http = "http";
my $remoteport = "80";

my $TIMEOUT = "timeout";
my $FAILURE = "Failure";
my $COMM_FAILURE = "CommunicationsFailure";
my $SUCCESS = "Success";
my $COMMAND = "command";

my $QueryDBID;     # Query id used for CreateNewQuery, ExecuteQuery
my @colNames;      # Names of columns
my $colHeaders;    # Number of columns
my $colPtr;        # Next column (field) being acquired for record

my @row;           # Query rows, once completed
my $rowTotal;      # Total number of rows
my $rowBatchCount; # Number of rows to load in a batch
my $rowPtr;        # Next row being read in

my $QueryMax;      # Maximum number of rows in a query
my $QueryColMax = 1000;  # Maximum column width in a query

# List of protocol tokens, mapping standard tokens to Java protocol tokens
my %token = (
    # Operators
    "OR"               => "0",
    "AND"              => "1",

    "EQ"               => "1",
    "NEQ"              => "2",
    "LT"               => "3",
    "LTE"              => "4",
    "GT"               => "5",
    "GTE"              => "6",
    "LIKE"             => "7",
    "NOT_LIKE"         => "8",
    "BETWEEN"          => "9",
    "NOT_BETWEEN"      => "10",
    "NULL"             => "11",
    "NOT_NULL"         => "12",
    "IN"               => "13",
    "NOT_IN"           => "14",
    );

my ($csetSchema, $cvobSchema); 
my (%crChangeSets, %csetVobs, %crDbIDs, %vobDbids);
my $cachePreloaded = 0;

my $batchProcessing = 0;
my %batchCommandHash = (); 

######################################################################
#::function CCcommand

=over 6

=item C<CCcommand(keyList)>

Encodes pairs to a CQWebJava command string and encodes strings so that
"command"=>"ASS OC" produces "command=ASS%20OC" with &'s in between Even
though the arguments represent hash pairs they will be processed as a
list to maintain the original order

If the key in the pair is "" then it is handled as a raw "append"
request with the value being appended to the command without being
touched. This allows a precomputed string like basic_cmd to be added
on at the end.

=back

=cut

######################################################################

sub CCcommand {
   my @pairs = @_;
   my $result="";
   my $debugResult = "CCcommand: ";
   my $command = "";

   # Assume first pair is command
   if ($pairs[0] eq "command") {
      $command = $pairs[1];
   }
   while (@pairs) {
       my ($op, $value) = splice(@pairs,0, 2);
       my $escapedValue;
       # Pre-CC7 clients will effectively perform an ASCII-check
       # CC7 clients will do code page conversion
       my $valueUTF8 = _convertValue("command $command $op",
           $value,"UTF8_CODEPAGE");
       if ($] ge "5.008006") {
           # uri_escape_utf8 is only available in 5.8.6 and beyond
           eval {
               $escapedValue = URI::Escape::uri_escape_utf8($valueUTF8);
           };
       } else {
           $escapedValue = uri_escape($valueUTF8);
       }
        $result .= ($result ? "&" : "") . 
                   ($op ? "$op=" . $escapedValue : $value);
       if ($op eq "pwd") {
           $value = "XXXXXX";
       }
       $debugResult .= ($op ? "$op=$value" : $value) . "\n";
   }
   &IODebugPrint("DETAILED", "$debugResult\n");
   return $result;
}

sub maskpw {
    my ($cmd) = @_;
    $cmd =~ s/\&pwd=(.*?)\&/\&pwd=XXXXXX\&/g;
    return $cmd;
}

######################################################################
#::function ClassInit

=over 6

=item C<ClassInit()>

This class function is automatically invoked to define class 
parameters using CQCC::Parm objects. Current parameters include: 

=item C<CQCC_URL>

The full URL needed to locate the ClearQuest web server. If not
provided directly it will be built up from the CQCC_SERVER and 
CQCC_SERVERROOT class parameters. 

=item C<CQCC_SERVER>

Name of the ClearQuest web server host. Used to build up the CQCC_URL
string along with CQCC_SERVERROOT.

=item C<CQCC_SERVERROOT>

Main folder on web server where server will respond to API calls.
Used to build up the CQCC_URL string along with CQCC_SERVER.

=item C<CQCC_WEB_DATABASE_SET>

Similar to CQCC_DATABASE_SET but only used by the CQWeb interface.
May or may not be set to the same value. No default value.

=item C<CQWEB_KEEPSESSION>

The CQWeb interface performance is greatly improved when the same
communications session is kept open and reused as long as it can be
(CQWEB_SESSION_TIMEOUT). This parameter enables reusing sessions and
is TRUE by default.

=cut

######################################################################

sub ClassInit { 
    if (&ParmDefined("CQCC_WEB_QUERY_MAX")) {
        return;
    }
    &SetParmDefaults(PCLASS=>"CQCC::$className", SRC_EV=>1, SRC_CFG=>1);
    &DefineParm(NAME=>"CQCC_SERVER_SSL", DEFAULT=>"FALSE", 
                RANGE_LIST => ["FALSE", "TRUE"]);

    &DefineParm(NAME=>"CQCC_URL");

    &DefineParm(NAME=>"CQCC_SERVER");

    &DefineParm(NAME=>"CQCC_SERVERROOT", DEFAULT=>"cqweb" );

    &DefineParm(NAME=>"CQCC_WEB_DATABASE_SET");

    &DefineParm(NAME=>"CQWEB_KEEPSESSION", DEFAULT=>'TRUE',
                RANGE_LIST => ["FALSE", "TRUE"]);

    # Internal value, used in testing only     
    &DefineParm(NAME=>"CQCC_WEB_QUERY_MAX", DEFAULT=>100,
                RANGE_MIN=>10, RANGE_MAX=>1000);
    &SetParmDefaults();
}

&ClassInit();

######################################################################
#::function ClassTest

=over 6

=item C<ClassTest(@args)>

This class function uses sets up a local environment for testing;
calls IOClassFactory() to get an instance of the desired subclass; 
and invokes the object's SelfTest method for testing.

=back

=cut

######################################################################

sub ClassTest { 
    my @args = @_;

    &ForceParm("ClearQuestInterface", "CQCC::$className");

    my $a = "Test encoding ampersand(&), less-than(<), greater-than(>), " .
            "quote(\"), and apostrophe (\')";
    print "Original string: $a\n";
    my $b = &XMLencode($a);
    print "Encoded string: $b\n";
    my $c = &XMLdecode($b);
    print "Decoded string: $c\n";

    my $cd = "]]>This is a ]]> test of the system]]>";
    print "original: $cd\n";
    print "CCDATAencode: " . &CDATAencode($cd) . "\n";

    return &CQCC::CQuest::ClassTest(@args);
}

######################################################################
#::method ClearCookie 

=over 6

=item C<ClearCookie()>

This method clears the persistent session cookie and timestamp, and 
deletes the session id file so future sessions won't attempt to use
it. 

=back

=cut

######################################################################

sub ClearCookie {
    my $self = shift;

    if (-r $self->{SESSIONFILE}) {
        unlink $self->{SESSIONFILE};
    }
    $self->{COOKIE}='';
    $self->{COOKIE_DB}='';
    $self->{TIMESTAMP}=0;

    # Split the stored cookie into multiple fields used for CQWebJava
    $self->{JAVA_COOKIE} = '';
    $self->{JAVA_SESSIONID} = '';
    $self->{JAVA_TIMEOUT} = '';
}

######################################################################
#::method ExecCommand

=over 6

=item C<($result, $operation, $error, $errorCode) = ExecCommand(cmdlist)>

This method opens up a fresh communications pipeline to ClearQuest
for each command; uses the Cookie management routines to restore or
save the session information between trigger invocations; sends the
command to ClearQuest and processes and packages the output for 
consumption by PostCommand(), its' only expected caller.

Returns $result (Success, Failure, Timeout, CommunicationsFailure)
        $operation (current command)
        $error (message)
        $errorCode (number)
=back

=cut 

######################################################################

sub ExecCommand {
    my ($self, @origCmdArgs) = @_;  # Argument passed in

#:::Setup socket
    ##############################
    # Socket setup
    my $sockaddr4 = 'S n a4 x8';
    my ($name, $aliases, $proto) = getprotobyname('tcp');
    my $family;
	
#:::Write to the server
    # Connect directly to server indicated by URL
    my $remotehost;
    my $url = $self->{URL};
    if ($url =~ m|^$http://(.*)|) {
       $remotehost = $1;
    }

    my $request = "/";    # default
    ($remotehost =~ s|^([^/]+)/(.*)$|$1|) && ($request = $2);
    ($remotehost =~ s/:(\d+)$//) && ($remoteport = $1);
    &IODebugPrint("DETAILED", "Host is $remotehost:$remoteport\n");

    my $there;
    if (IPV6_IS_SUPPORTED) {
        my ($af, $socktype, $proto, $saddr, $canonname, @res)
            = Socket6::getaddrinfo($remotehost, $remoteport, AF_UNSPEC, SOCK_STREAM);
        $there = $saddr;
        $family = $af;
		
    } else {
        $family = AF_INET;
        my $thataddr;
        if ($] ge "5.006") {
            $thataddr = inet_aton($remotehost);
        } else {
            my ($name, $aliases, $type, $len, $addr) 
                = gethostbyname($remotehost);
            $thataddr = $addr;
        }
        $there = pack($sockaddr4, $family, $remoteport, $thataddr);
    }

    # Remove the hostname from the URL
    if (!($url =~ s|$http://$remotehost:$remoteport||)) {
        $url =~ s|$http://$remotehost||;
    }

    my ($result, $error, $errorCode);
    my $done = 0;
    my $tries = 0;

    $self->LoadCookie();

    # If command fails because of timeout then go around second time
    # and create new session and login again then try command. 

    while (!$done && $tries < 2) {
        $tries++;

        # Create new session if existing cookie no longer valid
        if ($self->{COOKIE} eq '') {
            ($result, $error, $errorCode) = 
                $self->_ProcessMessage($url, $proto, $family, $there, 
                                       $remotehost, 1, 
                                       "command=CreateSession");
            if ($result ne $SUCCESS) {
                return ($result, "CreateSession", $error, $errorCode);
            }
            $self->{COOKIE_DB} = "";  # Clear cookie db to force login
         }

         # Login if not already logged into the right database
         # Fail if wrong login info
         # Expect to fail during initial login or changing databases
         # otherwise we expect to have the most recent database user/pw 
         # information still cached. 
         
         if ($self->{COOKIE_DB} ne &GetParm("CQCC_DATABASE")) {
             my $login = &CCcommand(
                 "command" => "Login",
                 "rmsessionid" => $self->{JAVA_SESSIONID},
                 "userid" => $self->{USER},
                 "pwd"    => $self->{PASS},
                 "userDb" => &GetParm("CQCC_DATABASE"),
                 "schemaRepo" => &GetParm("CQCC_WEB_DATABASE_SET"),
                 "displayName" => $self->{IP_ADDR},
                 "hostName" => $self->{LOCALHOST},
                 "restrictedModeInfo" => "false");

             ($result, $error, $errorCode) = 
                 $self->_ProcessMessage($url, $proto, $family, $there, 
                                        $remotehost, 0, $login);
                
             if ($result eq $TIMEOUT && $tries < 2) {
                 next;
              
             } elsif ($result ne $SUCCESS) {
                my %cmdArgs = @origCmdArgs;  # Converts pairs list to assoclist
                # The NOOP command is used in login attempts
                # So only try a new login if not already in one
                if ($cmdArgs{"command"} ne "NOOP") {
                    $self->Connect();
                    next;
                }
                return ($result, "Login", $error, $errorCode);
             }
             $done = 1;
        }

        my @commandArgs = @origCmdArgs;
        if ($self->{JAVA_SESSIONID} ne '') { 
            push(@commandArgs, "rmsessionid" => $self->{JAVA_SESSIONID});
        }
        my $cmdstring = CCcommand( @commandArgs );

        ($result, $error, $errorCode) = 
            $self->_ProcessMessage($url, $proto, $family, $there, 
                                   $remotehost, 0, $cmdstring);
                
        if ($result eq $TIMEOUT) {
            next;
        }
        $done = 1;
    }

#:::Announce results, clear or save cookie as needed
    &IODebugPrint("DETAILED",
        "session result: $result " . 
        "Cookie: $self->{COOKIE}\n");

    if ($result eq $COMM_FAILURE) { 
        # if no result is read, could be bad EVs, connection down, 
        #     server misconfigured, sent bad command, etc.

        &IOMessage("ERROR", ["CQWJ_E_POSTCOMMAND", $self->{URL}]);
        $self->ClearCookie();
    } else {
        $self->SaveCookie();  # Save to file and reset timestamp
    }

#:::Return value
    return ($result, $COMMAND, $error, $errorCode);
}

######################################################################
#::method _ProcessMessage

=over 6

=item C<($result, $error, $errorCode) = _ProcessMessage(message)>

Internal subroutine used to send individual command and process the 
results. 

=back

=cut

######################################################################

sub _ProcessMessage {
    my ($self, $url, $proto, $family, $there, $remotehost,
        $newSession, $message) = @_;

    #INTERNAL TESTING ONLY - Used to simulate problems making (dis)associations
    if ($ENV{"CQWEB_FAKE_ERROR"} && $message =~ /ssociate/) {
        print "CQWebJava::ProcessMsg \n$message\n" . 
            "Enter 'e' to simulate an error:";
        my $answer = <STDIN>;
        chomp $answer;
        if ($answer eq "e") {
            return ($FAILURE, "Faked error", "0");
        }
    }

    if ($self->{XML_PARSER}) {
        return _ProcessMessageXML(@_);
    }    

    my ($result, $error, $errorCode);
    my $buffer;
    my $endHeader = 0;
    my $insideQuery = 0;
    my $insideError = 0;
    my @rowColumns; 
    my $ldapEnabled =0;

    my $contentlen = length($message);
    # String sent to http server
    my $toserver = "POST $url HTTP/1.0\n" .
        "Content-Length: $contentlen\n" .
        "Content-Type: application/x-www-form-urlencoded\n";

    # If no cookie then load it from session file
    # Check the time and clear any stale cookies

    if (! $newSession && $self->{JAVA_COOKIE}) {
        $toserver .= "Cookie: $self->{JAVA_COOKIE}\n";
    }

    # the \n is needed to separate the header!
    $toserver .= "\n$message";

#:::: Send the request
    &IODebugPrint("DETAILED", "SENDING TO SERVER:\n" . &maskpw($toserver) . "\n");

    if ($ENV{"CQCC_UTIL_DEBUG"} > 2) {
        my $time = localtime(time);
        &IOFileAppend($self->{JAVA_COOKIE_LOG},
           "$time: $$, cookie: $self->{COOKIE} sessionid: $self->{JAVA_SESSIONID} " . 
           "java_cookie: $self->{JAVA_COOKIE} series=$ENV{CLEARCASE_SERIES_ID}\n");
    }
    socket(RSVR, $family, SOCK_STREAM, $proto) or 
        return($COMM_FAILURE, "Can't create server socket: $!");
    connect(RSVR, $there) or 
        return($COMM_FAILURE, "Can't connect to $remotehost: $!");
    select(RSVR); $| = 1; select(STDOUT);

    print RSVR $toserver;

#::::Get the content
    $result = $FAILURE;
    $error = "";
    $errorCode = "";

    &IODebugPrint("DETAILED", 
                   "\n\n******************** Read from server:\n\n");

    my $inputBuffer;
    while(read(RSVR,$inputBuffer,8*1024)) {
        $buffer .= $inputBuffer;
    }
    close RSVR;

    if ($buffer) { 
        my @lines = split(/^/, $buffer);
        my $line;

        foreach $line (@lines) {
            chomp $line;
            &IODebugPrint("DETAILED", "$line");

            # _ProcessMessage is only called if pre-CC7 client so this
            # call will effectively be an ASCII check only if enabled. 

            $line = &_convertValue("CQWebJava Response",$line,"CORE_CODEPAGE");

            ########################################
            # Header parsing

            if (!$endHeader) {
                if ($line =~ m/^Set-Cookie: (.*);\s*Path/) {
                    # Set-Cookie is good when creating a new session
                    if ($newSession eq 1) {
                        # save cookie, flush later if session gone
                        $self->{JAVA_COOKIE} = $1;
                        &IODebugPrint("DETAILED",
                            "Cookie set: $self->{JAVA_COOKIE}\n");
                    # Otherwise it indicates a timeout error
                    } else {
                        $result = $TIMEOUT;
                        last;
                    }

                # Key off of the double newline at end of header
                } elsif (length($line) <= 1) {
                    $endHeader = 1;
                }
             }
                
            ########################################
            # Body parsing
            
            if ($endHeader) {
                # General body message components
                if ($insideQuery eq 0) {
                    # All 
                    if ($line =~ m/\<Response [\S|\s]* Status='(\S*)'/) {
                        $result = $1;
                        if ($result eq $FAILURE) {
                            if ($line =~ m/ErrorCode=\'12000\'/) {
                                $result = $TIMEOUT;
                            }
                            if ($line =~ m/ErrorCode=\'(\S*)\'/) {
                                $errorCode = $1;
                            }
                        }
                    }

                    # Error Message
                    if ($line =~ m/\<Error /) {
                        $insideError = 1;
                    }
                    if ($insideError ne 0) {
                        $error .= "$line\n";
                        if ($line =~ m/\<\/Error\>/) {
                            $insideError = 0;
                        }
                    }

                    # CreateSession 
                    if ($line =~ m/\<Session ID='(\S*)'/) {
                        $self->{JAVA_SESSIONID} = $1;
                    } 
 
                    # If LDAP Authentication is enabled, set $ldapEnabled
                    if ($line =~ m/\<LDAP\S*\>(\w+)\<\/LDAP\S*\>/ ) {
                        &IODebugPrint("DETAILED","LDAPenabled. \n") ;
                        $ldapEnabled = ($1 eq "TRUE") ? 1 : 0;
                    }

                    # get CQ username from LDAP authenticaton response
                    if ($line =~ m/\<UserLoginName\>(\w+)\<\/UserLoginName\>/) {
                       if ($ldapEnabled eq "1") {
                           my $cqusername = $1;
                           # save cqusername to paramfile
                           $self->SetCQUser($cqusername);
                           &IODebugPrint("DETAILED","CQUSER = $cqusername \n");
                       }
                    }

                    # CreateNewQuery
                    if ($line =~ m/\<QueryDef [\S|\s]* dbid='(\S*)'/) {
                        $QueryDBID = $1;
                    }

                    # ExecuteQuery
                    if ($line =~ m/\<Resultset/) {
                       $insideQuery = 1;
                    }

                # Query response body message components
                } else {
                    # Start column definitions, get count, initialize
                    if ($line =~ m/\<ColumnHeaders Count='(\S*)'/) {
                        $colHeaders = $1;
                        $colPtr = 0;
                        &IODebugPrint("DETAILED","ColumnHeaders: $colHeaders");

                    # Get Column Names
                    } elsif ($line =~ m/\<ColumnHeader [\S|\s]*>(\S*)\</) {
                        $colNames[$colPtr] = $1;
                        &IODebugPrint("DETAILED","colNames[$colPtr] = " . 
                                      $colNames[$colPtr]);
                        $colPtr++;

                    # Get Row totals
                    } elsif ($line =~ m/\<Rows TotalCount='(\S*)' Count='(\S*)' StartRow='(\S*)'/) {
                        $rowTotal = $1;
                        $rowBatchCount = $2;
                        $rowPtr = $3 - 1;
                        &IODebugPrint("DETAILED", 
                            "rowTotal: $rowTotal " .
                            "rowBatchCount: $rowBatchCount " .
                            "rowPtr: $rowPtr");

                    } elsif ($line =~ m/\<Row dbid='(\S*)'/) {
                        my $dbid = $1;
                        if (!defined $dbid || $dbid eq '') {
                            $dbid = "(dbid?)";
                        }                        
                        @rowColumns = ($dbid);
                    
                    } elsif ($line =~ m/\<Column [\S|\s]*\>([\S|\s]*)\</) {
                        my $data = &XMLdecode($1);
                        push(@rowColumns, $data);
                    } elsif ($line =~ m/\<\/Row>/) {
                        my @localRow = @rowColumns;
                        $row[$rowPtr] = \@localRow;
                        &IODebugPrint("DETAILED", "row[$rowPtr]: " . 
                                      join(",", @{$row[$rowPtr]}) . "\n");
                        $rowPtr++;

                    } elsif ($line =~ m/\<\/ResultSet\>/) {
                        $insideQuery = 0;
                    }
                }
            } # end body-parsing
            # handle parsing for $error and $warning values
        } # end foreach-line
    } # end while-read-RSVR

    # for stale cookie, unlink the cookie file so the error won't
    # occur again
    if ($result eq $TIMEOUT || $result eq $FAILURE) {
        $self->ClearCookie();
    }
    return ($result, $error, $errorCode);
}

######################################################################
#::method _ProcessMessageXML

=over 6

=item C<($result, $error, $errorCode) = _ProcessMessageXML(message)>

Internal subroutine used to send individual command and process the 
results that uses the XML SAX parser support and SSL communications
sockets. 

=back

=cut

######################################################################

sub _ProcessMessageXML {
    my ($self, $url, $proto, $family, $there, $remotehost,
        $newSession, $message) = @_;
    my $response = "";
    my ($result, $error, $errorCode) = ($FAILURE,"","");
    my $ldapEnabled =0;
    my $cookie;

    ########################################
    # Build message to send out

    # If no cookie then load it from session file
    # Check the time and clear any stale cookies

    if (! $newSession && $self->{JAVA_COOKIE}) {
        $cookie = $self->{JAVA_COOKIE};
    }

    # TRY to send the message and parse the result
    eval {
        require CQCC::CQWebJavaParser;
        require HTTP::Response;

        ########################################
        # Send the request to the XML parser class
    
        # CQWEBJAVA_LWP_USERAGENT is a testing back door
        if ($http eq "https" || $ENV{CQWEBJAVA_LWP_USERAGENT}) {
            &IODebugPrint("DETAILED", "SENDING TO SERVER:\n" . &maskpw($message) . "\n");
        
            my $messageUTF8 = &_convertValue("command",$message,"UTF8_CODEPAGE");

            my $http_response = CQCC::CQWebJavaParser::ParserSendMessage(
                $http, $remotehost, $remoteport, $url, $cookie, $messageUTF8);

            # Process header information

            my $http_cookie = $http_response->header("Set-Cookie");
            my $http_content_type = $http_response->header("Content-Type");
            if ($http_cookie) {
                $http_cookie =~ s/;\s*Path=\S*//;
                # Set-Cookie is good when creating a new session
                if ($newSession eq 1) {
                    # save cookie, flush later if session gone
                    $self->{JAVA_COOKIE} = $http_cookie;
                    &IODebugPrint("DETAILED",
                           "Cookie set: $self->{JAVA_COOKIE}\n");
                    # Otherwise it indicates a timeout error
                } else {
                    $result = $TIMEOUT;
                    return ($result,$error,$errorCode);
                }
            }

            $response = $http_response->content();
            # Assume incoming response is utf8 but may not be marked properly
            eval {
                if (! Encode::is_utf8($response)) {
                    Encode::_utf8_on($response);
                }
            };
        } 

        ########################################
        # Else use simple socket connection

        else {
            my $buffer;

            # String sent to http server
            my $toserver = "POST $url HTTP/1.0\n" .
                "Content-Length: " . length($message) . "\n" .
                "Content-Type: application/x-www-form-urlencoded\n" . 
               ($cookie ? "Cookie: $self->{JAVA_COOKIE}\n" : "") .
               "\n$message";

            # Send the request
            &IODebugPrint("DETAILED", "SENDING TO SERVER:\n" . &maskpw($toserver) . "\n");

            # These returns will be caught by the eval and hit the
            # end of method code below. 

            if (! socket(RSVR, $family, SOCK_STREAM, $proto)) {
                $error = "Can't create server socket: $!";
                $result = $COMM_FAILURE;
                $!=undef;
                return($result, $error, $errorCode);
            }
            if (! connect(RSVR, $there)) {
                $error = "Can't connect to $remotehost: $!";
                $result = $COMM_FAILURE;
                $!=undef;
                return($result, $error, $errorCode);
            }
            select(RSVR); $| = 1; select(STDOUT);
            binmode(RSVR,":raw :utf8");
            print RSVR $toserver;

            # Retrieve the response
            my $endHeader = 0;

            my $inputBuffer;
            while(read(RSVR,$inputBuffer,8*1024)) {
                $buffer .= $inputBuffer;
            }
            close RSVR;

            if ($buffer) { 
                my @lines = split(/^/, $buffer);
                my $line;
                foreach $line (@lines) {
                    chomp $line;
                    
                    if (!$endHeader) {
                        if ($line =~ m/^Set-Cookie: (.*);\s*Path/) {
                            # Set-Cookie is good when creating a new session
                            if ($newSession eq 1) {
                                # save cookie, flush later if session gone
                                $self->{JAVA_COOKIE} = $1;
                                &IODebugPrint("DETAILED",
                                    "Cookie set: $self->{JAVA_COOKIE}\n");
                            # Otherwise it indicates a timeout error
                            } else {
                                $result = $TIMEOUT;
                                return ($result,$error,$errorCode);
                            }
                        # Key off of the double newline at end of header
                        } elsif (length($line) <= 1) {
                            $endHeader = 1;
                        }
                     }
                     elsif ($endHeader) { 
                         $response .= "$line\n";
                    }
                }
            }
        }

        ########################################
        # Retrieve the response

        &IODebugPrint("DETAILED", 
                       "\n\n******************** Read from server:\n\n" . 
                       "$response\n");

        ########################################
        # Parse the message
 
        # Response is assumed to be utf8 from input

        my ($r,$propsHashRef) = 
            CQCC::CQWebJavaParser::ParserProcessXML($response);
        my $elements = &_xmlProperty($propsHashRef, "Parser.ElementList");

        # Retrieve STATUS code (Success/Failure)
        $result = &_xmlProperty($propsHashRef, "Response.Status");

        # Retrieve ErrorCode and Error message if set
        #     ErrorCode='12000' means result = $TIMEOUT        
        if ($result ne $SUCCESS) {
            $errorCode = &_xmlProperty($propsHashRef, "Response.ErrorCode");
            if ($errorCode eq '12000') {
                $result = $TIMEOUT;
            }
            if ($elements =~ /Error/) {
                $error = &_xmlProperty($propsHashRef, "Error.Detail");
            }
            return ($result,$error,$errorCode);
        }

        if ($elements =~ /Session/) {
            $self->{JAVA_SESSIONID} = &_xmlProperty($propsHashRef, "Session.ID");

            # $ldapEnabled = LDAP value
            $ldapEnabled = &_xmlProperty($propsHashRef, "Session.LDAPAuthenticated");
            # if $ldapEabled use UserLoginName value
            if ($ldapEnabled eq "TRUE") {
                my $cqusername = &_xmlProperty($propsHashRef, "Session.UserLoginName");
                # save cqusername to paramfile
                $self->SetCQUser($cqusername);
                &IODebugPrint("DETAILED","CQUSER = $cqusername \n");
            }
        }

        if ($elements =~ /QueryDef/) {
            # $QueryDBID = QueryDef dbid
            $QueryDBID = &_xmlProperty($propsHashRef, "QueryDef.dbid");
        }

        if ($elements =~ /Resultset/) {
            # Retrieve ColumnHeaders and count
            $colHeaders = &_xmlProperty($propsHashRef, "ColumnHeaders.Count");

            # Retrieve ColumnNames - stored into first row        
            @colNames = &_xmlProperty($propsHashRef, "ColumnHeaders.ColNames");

            # Retrieve Row info - $rowTotal, $rowBatchCount, $rowPtr
            $rowTotal = &_xmlProperty($propsHashRef, "Rows.TotalCount");
            $rowBatchCount = &_xmlProperty($propsHashRef, "Rows.Count");
            # TRICKY - need to use this to APPEND new row data to existing rows
            $rowPtr = &_xmlProperty($propsHashRef, "Rows.StartRow");

            # Retrieve Query rows
            my %propsHash = %{ $propsHashRef };
            my $value = $propsHash{ "Resultset.Rows" };
            my @resultRows = @{ $value };
            my @queryRows;
            my $count = 1;
            for (@resultRows) {
                my @vals = &_convertValue("Row_$count",$_,"CORE_CODEPAGE");
                push @queryRows, \@vals;
                $count++;
            }

            push @row, @queryRows;
            $rowPtr = $#row + 1;
        }
        # Recover information about association and disassociation operations
        if ($elements =~ /Entity/) {
        }
    }; # end eval

    # CATCH errors from the TRY
    if ($@) {
        # TODO: Error attempting to instantiate class
        IOMessage("FATAL", "ERROR: CQWebJavaParser parser\n$@\n");
    }
    if ($result eq $TIMEOUT) {
        $self->ClearCookie();
    }
    return ($result, $error, $errorCode);
}

######################################################################
#::function _xmlProperty 
# Convert raw XML result value (utf8) to local codepage 
# Returns list or scalar depending on caller.

sub _xmlProperty {
    my ($propsHashRef, $name) = @_;
    my %propsHash = %{ $propsHashRef };
    my $value = $propsHash{ $name };
    return &_convertValue($name,$value,"CORE_CODEPAGE");
}

######################################################################
#::function _convertValue
# Accepts a scalar, arrayRef or arrayRef of arrayRefs
# Returns list or scalar depending on caller.

sub _convertValue {
    my ($name, $value, $dst_encoding) = @_;
    my ($status,@results);

    if ($value) {
        my $src_encoding = ($dst_encoding eq "UTF8_CODEPAGE") ? 
                           "CORE_CODEPAGE" : "UTF8_CODEPAGE";
        ($status,@results) = 
            &IOConvertString($value, $src_encoding, $dst_encoding);
        if ($status) {
            &IOMessage("FATAL", ["CQWJ_E_CONVERSION_ERROR", $name, $status]);
        }
    }
    if (wantarray()) {
        return @results;
    }
    return $results[0];
}

######################################################################
#::method LoadCookie

=over 6

=item C<LoadCookie()>

This subroutine will load or validate the persisten session "cookie".
If not keeping session then cookie should already be empty. 
If cookie not set then read it from session file if available.
If stale cookie, then clear it and return no cookie. 

=back

=cut

######################################################################

sub LoadCookie {
    my $self = shift;
    my ($age, $result, $dbset);

    # NOTE: If no file, it reports a stale cookie with big timeout
    # If no cookie then restore it from file if available

    if ($self->{COOKIE} eq '' && -r $self->{SESSIONFILE}) {
       # Note - stat/$mtime is not reliable as a timestamp from 
       # the current host since it relies on the filesystem where
       # the session file is saved which may have a different clock

       &IODebugPrint("DETAILED", "LoadCookie: Restoring from sessionfile\n");

       # Unless timestamp read in, will clear
       $self->{TIMESTAMP} = 0;  

       ($self->{COOKIE}, $self->{TIMESTAMP}, $self->{COOKIE_DB}) = 
           &IOFileRead($self->{SESSIONFILE},0);
    }

    # ExecCommand will automatically catch timeouts and recreate sessions
    # and relogin to the database. Don't try to catch stale cookies or 
    # bad databases. 

    $age = time() - $self->{TIMESTAMP};
    &IODebugPrint("DETAILED", "LoadCookie cookie, $age secs old\n");

    # Split the stored cookie into multiple fields used for CQWebJava
    ($self->{JAVA_COOKIE}, $self->{JAVA_SESSIONID}, $self->{JAVA_TIMEOUT},
        $dbset) =  split(/,/, $self->{COOKIE});
    if (! (defined($self->{JAVA_SESSIONID}) &&
           $dbset eq &GetParm("CQCC_WEB_DATABASE_SET"))) {
        $self->ClearCookie();
    }
}

######################################################################
#::method Logon

=over 6

=item C<Logon()>

This method attempts to actively logon to the target ClearQuest
database. Once established the class will try to use the same
session id as long as possible. 

=back

=cut

######################################################################

sub Logon {
    my $self = shift;

    # If somebody cleared the login parm file, make sure we clear old cookie
    if (! -s $self->{PARMFILE}) {
       $self->ClearCookie();
    } 

    $self->GetLogonParms();

    &IODebugPrint("BASIC",
        "Test Mode: Trying to connect to $self->{URL}\n" . 
        "ClearQuest User is: $self->{USER}\n" . 
        "Connecting to any available DB on remote system.\n\n");

    my $t1 = TimerStart($className . "::Logon");
    my $trace = (&GetParm("CQCC_DEBUG") > 1) ? "TRUE" : "FALSE";
    if ($self->PostCommand("command"=>"NOOP")) 
    {
        &IODebugPrint("BASIC", $className,
            "\nTest Mode SUCCESSFUL: communication succeeded with" .
            " remote server and CQ/CC integration server page.\n");
        TimerStop($t1);
        return 1;

    } else {
        &IODebugPrint("BASIC", $className,
            "\nTest Mode FAILURE: communication not successful to" .
            " remote server, OR to the integration server page.\n");
        $self->ClearCookie();
        TimerStop($t1);
    }

    return 0;
}

######################################################################
#::method new

=over 6

=item C<new CQCC::CQWebJava()>

This method initializes the object's attributes including the 
URL, name of the session file to store session information between
trigger invocations and the session id (cookie) used. 

=back

=cut

######################################################################

sub new { 
   my $proto = shift;
   my $class = ref($proto) || $proto;  # Object or class name

   # Attributes
   my $self;

   $http = (&GetParm("CQCC_SERVER_SSL") eq "TRUE") ? "https" : "http";

   # CCperl doesn't support SUPER so use direct reference to parent
   $self = new CQCC::CQuest();

   $self->{CLASSNAME} = "CQCC::$className";
   # In CQWebJava, the "cookie" loaded, saved, and cleared by CQWeb
   #    is three values that CQWebJava uses 


   # Don't rely on session file age, use embedded timestamp from client
   # Session File is a managed resource - it can be aqcuired and reused 
   # by any new thread in the same session and is released for reuse
   # on normal exit. 

   $self->{SESSIONFILE} =   # File containing sessionid, ~/.cqsession
       &IOResourceManager(".cqsession", 1, &GetParm("CM_CQSESSION_STALE_MINS")); 
   $self->{COOKIE} = "";    # CQ sessionid returned by web server
                            # Stored into ~/.cqsession file 
   $self->{COOKIE_DB} = "?"; # CQ database associated with cookie
   $self->{JAVA_SESSIONID} = ''; # CQWeb Java session id
   $self->{JAVA_COOKIE} = '';    # CQWeb Java cookie
   $self->{JAVA_TIMEOUT} = 0;    # CQWeb Java timeout information
   $self->{JAVA_COOKIE_LOG} = &IOPathname("PARM","CQWebJava.log");

   my $schemaRepo = &GetParm("CQCC_WEB_DATABASE_SET");   
   if ($schemaRepo eq "") {
      &IOMessage("FATAL",["CQWJ_E_DATABASE_SET_UNSET","CQCC_WEB_DATABASE_SET"]);
   } 

   # get the hostname and figure out IP address
   $self->{LOCALHOST} = hostname();
   $self->{IP_ADDR} = undef;

   if (IPV6_IS_SUPPORTED) {
       my ($family, $socktype, $proto, $saddr, $canonname, @res)
           = Socket6::getaddrinfo($self->{LOCALHOST}, '', AF_UNSPEC, SOCK_STREAM);
       ($self->{IP_ADDR}, my $port) 
           = Socket6::getnameinfo($saddr, Socket6::NI_NUMERICHOST | Socket6::NI_NUMERICSERV);
   } else {
       my ($hostname,$aliases,$addrtype,$length,@addrs) 
           = gethostbyname($self->{LOCALHOST});
       $self->{IP_ADDR} = inet_ntoa($addrs[0]);
   }
 
   # Use URL if provided else build up from components

   $self->{URL} = &GetParm("CQCC_URL"); # URL to CQ web database
   if (! defined $self->{URL}) {
       if (&GetParm("CQCC_SERVER") && &GetParm("CQCC_SERVERROOT")) {
           $self->{URL} = "$http://" . &GetParm("CQCC_SERVER") . 
               "\/" . &GetParm("CQCC_SERVERROOT") . "/cqwebapi";
       }
   }
   # Error if still not defined
   if (! defined $self->{URL}) {
       &IOMessage("FATAL", ["CQWJ_E_INVALID_SERVER"]);
   }

   # Set the XML parser to undefined
   $self->{XML_PARSER} = undef;
   if ($] ge "5.008") {
       eval {
           require CQCC::CQWebJavaParser;
           $self->{XML_PARSER} = CQCC::CQWebJavaParser::ParserInit();
           # Will need to Eval/Test all calls like this ? 
       };
       if ($@) {
           # TODO: Error attempting to instantiate class
           IOMessage("FATAL", "ERROR: Loading CQWebJavaParser\n$@\n");
       } 
   }

   # Set up the schema needed for handling secondary entities. 

   if (! defined $csetSchema) { 
       $csetSchema = 
          new CQCC::CQSchema(NAME=>"cc_change_set",
                             TYPE=>"cc_change_set",
                             ENTITY_LIST => "cc_change_set");
       $cvobSchema = 
          new CQCC::CQSchema(NAME=>"cc_vob_object",
                             TYPE=>"cc_vob_object",
                             ENTITY_LIST => "cc_vob_object");

       $cvobSchema->ChangeFieldMap(
            "VOBID" => "vob_family_uuid",
            "OID" => "object_oid",
            "NAME" => "name",
            "MODIFY" => "modify" );
   }

   bless($self, $class);
   return $self;
}

######################################################################
#::method PostCommand

=over 6

=item C<PostCommand($cmd)>

This method completes preparation of the command being sent, sends it
to ClearQuest using the ExecCommand() method, then interprets the
result codes returned from ClearQuest.

The return values are (0) - Failed; (1) Succeeded; and 2(Succeeded
and more data is available. 

=back

=cut

######################################################################

sub PostCommand {
    my ($self, @cmdArgs) = @_;
    my $post_result = 1;

    my $t1 = &TimerStart($className . "::PostCommand $cmdArgs[1]");
    &IODebugPrint("DETAILED", "PostCommand: $cmdArgs[1]\n");
    
    my ($result, $operation, $error, $errorCode) = 
        $self->ExecCommand(@cmdArgs);

    &IODebugPrint("DETAILED",
        "\n\n" . " returned status: $result errorCode: [$errorCode]\n" .
        "    operation: $operation\n" .
        "    error: $error\n");

    if ($error =~ "Database name is not valid") {
        &IOMessage("ERROR",["CQWJ_E_DATABASE_INVALID", $error]);
    }

    if ($error =~ /Code Page Mismatch/g) {
        &IOMessage("FATAL", [ "CQP_E_CODE_PAGE_MISMATCH"]);
    }

    if ($result eq $COMM_FAILURE) { 
        &IOMessage("FATAL", ["CQW_E_POSTCOMMAND", $self->{URL}]);
    }

    if ($result ne $SUCCESS) {
        my $summary = $error;
        if ($error =~ /\<Summary\>([\S\s]*)\<\/Summary\>/s) {
            $summary = $1;
        }
        &IOMessage("ERROR",["CQWJ_E_OP_FAILED",
                   (($operation eq $COMMAND) ? $cmdArgs[1] : $operation),
                   "($errorCode) $summary"]);
        $post_result = 0;
    }

    &TimerStop($t1);
    return $post_result;
}

######################################################################
#::method PrintChangeSets

=over 6

=item C<PrintChangeSets(dbName, entityName, @defectIds)>

This method prints out the cc_change_set and cc_vob_objects associated
with one or more defect records. 

=back

=cut

######################################################################

sub PrintChangeSets { 
    my ($self, $dbName, $entity, @defectIDs) = @_;

    my $id;
    my $schema = &GetCQSchemaByEntity($entity);

    $self->_LoadCSetCache($dbName,$schema, $entity, @defectIDs);

    # For each defect, locate it's changeset objects list, then 
    # get the cc_vob_objects that match each entry.

    for $id (@defectIDs) {
        my $csetDbID = $crChangeSets{ $id };
        &IOPrint("\n\nDefect $id cc_change_set list $csetDbID\n");

        if ($csetDbID == 0) {
            next;
        }

        # Process all the cc_vob_objects in the change set
        my $ccVobRef = $csetVobs{ $csetDbID };
        if (defined $ccVobRef) { 
            my @ccVobDbIDlist = @{ $ccVobRef };
            my @ccvDbIDList;

            # Gather list of cc_vob_object dbids for single query
            for (@ccVobDbIDlist) {
               my ($ccvDbID, $name) = split(" ", $_, 2);
               push @ccvDbIDList, $ccvDbID;
            }                
            my @ccVobList = 
                $self->PerformQuery($dbName, "cc_vob_object", 
                    $cvobSchema, 
                    [ "dbid in " . join(",", @ccvDbIDList) ],
                    [ "dbid,<NAME>,<VOBID>,<OID>", "%s=%s!%s!%s" ], 
                    "FALSE");

            for (@ccVobList) {
                IOPrint("\nCC_Vob_object: \n");
                my ($vDbID, $body) = split("=", $_);
                my ($vName, $vfUuid, $oid) = split("!", $body);
                &IOPrint("    Dbid: $vDbID\n" . 
                      "    Pathname: $vName\n" . 
                      "    Vob_Family_UUID: $vfUuid\n" . 
                      "    ObjectID: $oid\n");
            }
        }
    }
}

######################################################################
#::method ProcessBatchOps

=over 6

=item C<ProcessBatchOps(hash)>

This method processes the hash lists built up when the batch information
was processed from a file or other means. 

=back

=cut

######################################################################

sub ProcessBatchOps {
    my ($self,$cqHash) = @_;
    my %cqHash = %{$cqHash};
    my (%volHash, %addHash, %delHash, $key, $cqKey, %dbEntityHash);
    my $result = 1;

    $batchProcessing = 1;
    %batchCommandHash = ();

    # Break up the {defect}->({path}->(op,vobid,oid,oldOID),...) order
    # into several hash tables for easier processing by CQWebJava
    # Key based on "database,entity,pathname" to clump similar db-entity
    # sets together. 
    #    volHash:  {key}->"path,vobid,oid,oldOID"
    #    addHash:  {key}->"defect,defect,..."
    #    delHash:  {key}->"defect,defect,..."

    for $cqKey (sort keys %cqHash) {
        my ($entity,$dbID) = split(":",$cqKey);
        my $db = substr($dbID,0,length($dbID)-8);
        my $id = substr($dbID,-8);
        my %opHash = %{$cqHash{$cqKey}};
        my $path; 

        for $path (sort keys %opHash) {
            my ($op,$vobid,$oid,$oldOID) = split(",", $opHash{$path});
            $key = join(",", $db, $entity, $path);
            if (! defined $volHash{$key}) {
                $volHash{$key} = join(",",$path,$vobid,$oid,$oldOID);
                if ($op eq "<+>") {
                    $addHash{$key} = $id;
                } else {
                    $delHash{$key} = $id;
                }
            }
            else {
               if ($op eq "<+>") {
                   $addHash{$key} = $addHash{$key} . "," . $id;
               } else {
                   $delHash{$key} = $delHash{$key} . "," . $id;
               }
            }
            # Accumulate list of defects for this same db/entity 
            $key = join(",", $db, $entity);
            my $idKey = $db . $id;
            if (! defined $dbEntityHash{$key}) {
                my %idList = ($idKey => 1);
                $dbEntityHash{$key} = \%idList;
            } else {
                my %idList = %{$dbEntityHash{$key}};
                if (! defined $idList{$idKey}) {
                    $idList{$idKey} = 1;
                    $dbEntityHash{$key} = \%idList;
                }
            }
        }
    }
    my $addAssocs = new CQCC::AssocList();
    my $delAssocs = new CQCC::AssocList();

    my $currentDbEntity = "";
    for $key (sort keys %volHash) {
        my ($db, $entity, $path) = split(",", $key);
        my $voList = $volHash{$key};

        # Preload CSet Cache if not already loaded
        my $dbEntity = join(",", $db, $entity);
        if ($dbEntity ne $currentDbEntity) {
            my %idList = %{$dbEntityHash{$dbEntity}};
            my $init_err_count = &IOstd->{ERR_COUNT};
            $self->_LoadCSetCache($db, &GetCQSchemaByEntity($entity),
                                  $entity, (sort keys %idList));
            if (&IOstd->{ERR_COUNT} > $init_err_count) {
                $cachePreloaded = 0;
                $batchProcessing = 0;
                return 0;
            }
            $cachePreloaded = 1;
            $currentDbEntity = $dbEntity;
        }
          
         # Process this entity
         $self->SetDBEntity($db,$entity);
         $addAssocs->ProcessBugString($entity,$addHash{$key});
         $delAssocs->ProcessBugString($entity,$delHash{$key});
         my $r = $self->EditAssocs($voList, $addAssocs, $delAssocs);
         if ($r == 0) {
             $result = 0;
             $cachePreloaded = 0;
             $batchProcessing = 0;
             return $result;
         }
    }

    # Process any accumulated associate/disassociate commands
    for (sort keys %batchCommandHash) {
        my ($command, $entity, $entityDbID) = split(",", $_);
        my $schema = &GetCQSchemaByEntity($entity);

        my $r = $self->PostCommand(
            "command"=> $command,
            "dbid" => $entityDbID,
            "entityDefName" => $entity,

            #TODO: Next two were Associate only?
            "action" =>  $schema->{'FIELD_MAP'}->{'MODIFY'},
            "tabName" => "None",

            "fieldsXml" => $batchCommandHash{$_});
        if ($r == 0) {
            $result = 0;
            return $result;
        }
    }

    $cachePreloaded = 0;
    $batchProcessing = 0;
    return $result;
}

######################################################################
#::method SaveCookie 

=over 6

=item C<SaveCookie()>

This method will save the persistent session cookie along with a 
timestamp to ensure the session information is not used after it
has expired. 

=back

=cut

######################################################################

sub SaveCookie {
    my $self = shift;
    $self->{COOKIE} = join(",", $self->{JAVA_COOKIE}, $self->{JAVA_SESSIONID},
                    $self->{JAVA_TIMEOUT}, &GetParm("CQCC_WEB_DATABASE_SET"));

    # CQCC::CQWeb::SaveCookie($self);
    if ((&GetParm("CQWEB_KEEPSESSION") eq 'TRUE') && 
            ($self->{COOKIE} ne '')) {
        my $result;
        &IODebugPrint("DETAILED", 
            "SaveCookie cookie into sessionfile\n");
        $self->{TIMESTAMP} = time();
        $self->{COOKIE_DB} = &GetParm("CQCC_DATABASE");
        &IOFileWrite($self->{SESSIONFILE}, 
            $self->{COOKIE}, $self->{TIMESTAMP}, $self->{COOKIE_DB});
        # If unable to open sessionfile for writing, won't be able to
        # reuse sessionid on next trigger but will be ok on this one 
        # so don't report an error
    }
}

######################################################################
#::function tagElement

=over 6

=item C<tagElement(name, value)>

This function returns an XML tag entry for the value in the form:

<name>value</name>

=back

=cut

######################################################################

sub tagElement {
    my ($name, $value) = @_;
    if ($name =~ m/(\S*)\s/) {
        return "<$name>$value</$1>";
    }
    return "<$name>$value</$name>";
}

######################################################################
#::function token

=over 6

=item C<token(key)>

This function returns the appropriate token value for the current
protocol.

=back

=cut

######################################################################

sub token {
    my ($key) = @_;
    return $token{$key};
}

######################################################################
#::function CDATAencode

=over 6

=item C<encodedValue = CDATAencode($value)>

This function wraps a string inside a CDATA element which is required
for some CQWebJava commands such as Associate and Disassociate. The
CDATA element will be represented as "<![CDATA[string]]>". Since the
string may contain the closing marker "]]>" the routine looks for this
case and breaks such strings into multiple CDATA segments by splitting
in the middle of the "]]>" marker so it is broken across segments. 

=back

=cut

######################################################################

sub CDATAencode {
    my ($data) = @_;

    my $startToken = "<![CDATA[";
    my $endToken = "]]>";
    my $result = "";

    $data =~ s/\]\]\>/\]\]$endToken$startToken\>/g;
 
    return "$startToken$data$endToken";
}

######################################################################
#::function XMLdecode

=over 6

=item C<decodedValue = XMLdecode(value)>

Internal function returns a version of the input string in which any 
special XML characters that are character-entity encoded are replaced
with their original character representation.

=back

=cut

######################################################################

sub XMLdecode {
    my ($data) = @_;
    
    $data =~ s/\&amp\;/\&/g;
    $data =~ s/\&lt\;/\</g;
    $data =~ s/\&gt\;/\>/g;
    $data =~ s/\&quot\;/\"/g;
    $data =~ s/\&apos\;/\'/g;

    return $data;
}

######################################################################
#::function XMLencode

=over 6

=item C<encodedValue = XMLencode(value)>

Internal function returns a version of the input string in which the 5 
special XML characters have been replaced with their character-entity 
versions. 

=back

=cut

######################################################################

sub XMLencode {
    my ($data) = @_;
    
    $data =~ s/\&/\&amp\;/g;
    $data =~ s/\</\&lt\;/g;
    $data =~ s/\>/\&gt\;/g;
    $data =~ s/\"/\&quot\;/g;
    $data =~ s/\'/\&apos\;/g;

    return $data;
}

######################################################################
#::method _EditAssocs

=over 6

=item C<_EditAssocs(database,entity,$VOList,add,del)>

This internal method implements the EditAssocs functionality for
a single database and entity. The add and del association lists are
expected to be homogeneous since we can only interact with one
database and one entity type at a time. The EditAssocs method is
responsible for splitting up associations into homogeneous lists. 

It must be called thru the CQuest method, EditAssocs which will 
break up the association lists into unique database/entity 
combinations so that this method only handles one combination at a 
time. 

This method calls _LoadCSetCache to speed up getting dbids for 
cc_change_set, cc_vob_object entities as well as the OIDS when
required for the files.  This may be overkill in
some situations but should scale up well for larger numbers of 
associations per changeset.

=back

=cut

######################################################################

sub _EditAssocs {
    my ($self,$dbName,$entity,$VOList,$addAssocs,$delAssocs) = @_;
    my ($cqname, $vobid, $oid, $oldOID) = split(",", $VOList);
    my $result = 1;
    my $parmSet = 0;
    my (@addIDs, @delIDs, $id, $hashOID, $dupOID, $vobs, $delimiter);
    my $schema = &GetCQSchemaByEntity($entity);
    my %deletedKeys;  # Hash of elements deleted already

    # For efficiency and scalability, we do cache up the results of
    # two queries as hash tables to get the dbids we need for 
    # cc_change_sets and cc_vob_objects.     Sometimes a third
    # query is needed, this query has the OIDS and under certain
    # conditions we need to match the OIDS.

    if ($addAssocs) {
        @addIDs = $addAssocs->GetBugList();
        if (@addIDs && ($parmSet == 0)) {
            if (&GetParm("CQCC_QUERY_OID") eq "TRUE") {
                $parmSet = 1;
            }
        }
    } 
    if ($delAssocs) {
        @delIDs = $delAssocs->GetBugList();
        if (@delIDs && ($parmSet == 0)) {
            &ForceParm("CQCC_QUERY_OID", "TRUE");
            $parmSet = 1;
        }
    } 

    if ($cachePreloaded == 0) {
        if (&GetParm("CQCC_DEBUG") > 1) { 
            &IODebugPrint("DETAILED", "$className::_EditAssocs:\n" . 
                          "   Add: " . join(",",@addIDs) . "\n" . 
                          "   Del: " . join(",",@delIDs) . "\n" . 
                          "   _LoadCSetCache to get cc_change_sets\n");
        }
        my $init_err_count = &IOstd->{ERR_COUNT};
        $self->_LoadCSetCache($dbName,$schema, $entity, @addIDs, @delIDs);
        if (&IOstd->{ERR_COUNT} > $init_err_count) {
            return 0;
        }
    }

    $delimiter = $ENV{"CLEARCASE_XN_SFX"};
    if (! defined $delimiter) {
        $delimiter = "@@";
    }

    # Delete existing association if found
    # If no change_set record then ignore
    # Loop over delIDs for each defect
    #   delete assoc from objects list 
    #   update hash table

    if (@delIDs) {
        # Because we may be deleting checkout associations during
        # checkin, we might be working with $cqname being an XPN
        # and having to delete the checkout names which used PN.
        # So if the the existing CQ assocation is a PN name, 
        # then check it against the PN part of $cqname only or we 
        # won't match. 

        # Note - @@ may not always be delimiter but can't always rely
        # on ClearCase env

        my ($cqname_PN) = split($delimiter, $cqname);

        for $id (@delIDs) {
            my $entityDbID = $crDbIDs{ $id };
            my $csetDbID = $crChangeSets{ $id };
            if ($entityDbID eq "0" || $csetDbID eq "0") {
                # Can't find change set to delete entries from
                next;
            } 
            my $vobs = $csetVobs{ $csetDbID };
            if ($vobs) {
                my @vobsList = @{ $vobs };
                my $vob;
                for $vob (@vobsList) {
                    my ($dbid,$p) = split(" ", $vob, 2);
                    my ($p_pn, $p_branch) = split($delimiter, $p);
                    $hashOID = $vobDbids  { $dbid};
                                                
                    # Note: Don't delete VOB record because of 
                    # possible use by other entries

                    if ((defined $p_branch && $p eq $cqname) ||
                        (! defined $p_branch && $p eq $cqname_PN) ||
                        ((&GetParm("CQCC_QUERY_OID") eq "TRUE") &&
                            ($hashOID eq $oldOID))) {

                        my ($vob_dbid) = split(" ",$vob);
                        my $deleteFields = 
                            &tagElement("Field",
                                &tagElement("Name", "cc_change_set") .
                                &tagElement("Value", $vob_dbid));
                        if ($batchProcessing == 0) {
                            my $r = $self->PostCommand(
                                "command"=>"DisAssociate",
                                "dbid" => $entityDbID,
                                "entityDefName" => $entity,
                                "fieldsXml" => $deleteFields);
                            if ($r == 0) {
                                $result = 0;
                                return $result;
                            }

                        # Save info for later processing
                        } else {
                            my $key = "DisAssociate,$entity,$entityDbID";
                            $batchCommandHash{$key} .= $deleteFields;
                        }

                        # Remember the deleted element
                        $deletedKeys{"$id:$vob"} = 1;
                    }
                }
            }
        }
    }
    if (@addIDs) {  
        #Make association if possible
        #    Create new association entity with name, vobid, 
        #    Loop over addlist for each defect
        #        add assoc to objects list and add to defect if new
        #        update hash tables (?)

        my $assocEntry = 
            &tagElement("Value",
                    &tagElement("VobObjectName", &CDATAencode($cqname)) .
                             &tagElement("ObjectOID", $oid) .
                             &tagElement("VobFamilyUUID", $vobid));

        for $id (@addIDs) {
            my $entityDbID = $crDbIDs{ $id };

            # Check for existing change set entry with exact same pathname
            # to avoid making duplicate entries
            my $entryExists = 0;
            my $csetDbID = $crChangeSets{ $id };

            #If existing change set, then review vob entries for match
            if (defined $entityDbID && $entityDbID ne "0" && 
                defined $csetDbID && $csetDbID ne "0") {
                my $vobs = $csetVobs{ $csetDbID };
                if ($vobs) {
                    my @vobsList = @{ $vobs };
                    my $vob;
                    # If name matches any existing entry, skip this entry
                    for $vob (@vobsList) {
                       my ($dbid,$p) = split(" ", $vob, 2);
                       if ($parmSet) {
                           $hashOID = $vobDbids  { $dbid};
                           if ($hashOID eq $oid) {
                               $entryExists = 1;
                               last;
                           }
                       }
                       # If path matches and entry wasn't deleted earlier 
                       if (($p eq $cqname) && (! $deletedKeys{"$id:$vob"})) {
                           $entryExists = 1;
                           last;
                       }
                    }
                }
            }

            # If no existing entry, add a new one to the change set
            if ($entryExists eq "0") {
                my $addFields = &tagElement("Field",
                                            &tagElement("Name","cc_change_set")
                                            . $assocEntry);
                if ($batchProcessing == 0) {
                    my $r = $self->PostCommand("command"=>"Associate",
                                               "dbid" => $entityDbID,
                                               "entityDefName" => $entity,
                                               "action" =>  $schema->{'FIELD_MAP'}->{'MODIFY'},
                                               "tabName" => "None",
                                               "fieldsXml" => $addFields);

                    if ($r == 0) {
                        $result = 0;
                        return $result;
                    }
                } else {
                    # Save info for later processing
                    my $key = "Associate,$entity,$entityDbID";
                    $batchCommandHash{$key} .= $addFields;
                }
            }
        }
    }

    return $result;
}

######################################################################
#::method _PerformQuery 

=over 6

=item C<_PerformQuery(db,entity,schema,query,results,results_format, returnType, queryType)>

This internal method is called by PerformQuery to actually query the
ClearQuest database for records matching the query description.

Queries are the only type of request where we may have to generate
more than one command to get the entire result back; if the query
status returns 'PARTIAL' we need to send a 'FETCH' command to obtain
more result rows. Also, session status remains ACTIVE if more rows are
to be fetched, regardless of the 'KeepSession' setting.

=back

=cut

######################################################################

sub _PerformQuery {
    my ($self, $dbname, $entity, $schema, $query, $resultsSet,
        $resultsFormat, $returnType, $queryType) = @_;
    my $forceDBID = 0;

    if ($queryType eq "CQ") { 
        &IOMessage("ERROR", ["CQ_E_OP_UNSUPPORTED", $className, 
                             "PerformQuery with Named queries"]);
        return;
    }

    if (! defined $QueryMax) {
        $QueryMax = &GetParm("CQCC_WEB_QUERY_MAX");
    } 

    my @clauseList = @{ $query };  # Must be an array reference
    my ($filterList, @filters);

    # PostCommand/CCcommand will do the URLencoding needed
    $filterList = "<FilterNode>" . 
                  &tagElement("BoolOp",token("AND")) .
                  "<FieldFilters>";
    for (@clauseList) {
        my ($subject, $verb, $args) = split(" ", $_, 3);
        my $argsFilter = "<Values>";
        my @argsList = split(",",$args);
        for (@argsList) {
            $argsFilter .= &tagElement("Value", &XMLencode($_));
        }
        $argsFilter .= "</Values>";
        $subject = &XMLencode($subject);
        my $filter = &tagElement("FieldFilter ID='$subject'",
                                 &tagElement("CompOp", token(uc($verb))) .
                                 &tagElement("Prompt","") .
                                 $argsFilter);
        $filterList .= $filter;
    }
    $filterList .= "</FieldFilters>" .
                   "</FilterNode>";

    my $resultsFields;

    my @resultNames = split(",", $resultsSet);
    $resultsFields = "<QueryFieldDefs>";

    # Always force first item to be the dbid
    if ($resultNames[0] ne "dbid") {
       for (@resultNames) {
           if ($_ eq "dbid") {
               # Should be an internal development error if it occurs
               &IOMessage("ERROR", ["CQWJ_E_QUERY_ORDER", $resultsSet]);
               return ();
           }
       } 
       unshift(@resultNames,"dbid");
       $forceDBID = 1;
    }
    for (@resultNames) { 
        $resultsFields .= 
            &tagElement("QueryFieldDef ID='" . $_ . "'",
                        &tagElement("Label", &XMLencode($_)) .
                        &tagElement("SortType", "0") .
                        &tagElement("SortOrder", "0"));
    }        
    $resultsFields .= "</QueryFieldDefs>";
   
    # Execute the Query command then Fetch the results
    my @addedArgs;
    @addedArgs = ("startRow" => 0, "rowCount" => 100);

    my $result = $self->PostCommand("command"=>"CreateNewQuery",
                                    "edname"=>$entity);
    if (! $result) {
        return ();
    } 

    $result = $self->PostCommand("command" => "SetDisplayFieldsAndFilter",
                                 "dbid" => $QueryDBID,
                                 "newDisplayFieldsXML" => $resultsFields,
                                 "filterTreeXML" => $filterList);
    if (! $result) {
        return ();
    }

    @row = ();
    $result = $self->PostCommand("command" => "ExecuteQuery",
                                "dbid" => $QueryDBID,
                                "startRow" => "1",
                                "rowCount" => $QueryMax,
                                "resultsetColWidth" => $QueryColMax);
    if (! $result) {
        return ();
    }

    # Handle overflow queries based on counts of unfetched records
    
    while ($rowPtr < $rowTotal) {
        my $rowCount = ($rowTotal - $rowPtr + 1);
        if ($QueryMax < $rowCount) {
            $rowCount = $QueryMax;
        }
        $result = $self->PostCommand("command" => "GetRecords",
                                     "dbid" => $QueryDBID,
                                     "startRow" => $rowPtr,
                                     "rowCount" => $rowCount,
                                     "resultsetColWidth" => $QueryColMax);
        if (! $result) {
            return ();
        }
    }

    # Map recovered rows to the desired output format
    my @bugdesc;
    my $n = 0;
    my $rowArray;

    if ($returnType eq "STRING") {
        for $rowArray (@row) {
            my @bugRow = @{$rowArray};
            # Remove leading dbid if we added it above
            if ($forceDBID ne 0) {
                shift @bugRow;
            }
            $bugdesc[$n++] = sprintf($resultsFormat, @bugRow);
        }
    } else {
        # Include column labels as first record in return
        $bugdesc[$n++] = \@colNames;
        for $rowArray (@row) {          
            my @bugRow = @{$rowArray};
            # Remove leading dbid if we added it above
            if ($forceDBID ne 0) {
                shift @bugRow;
            }
            # Only try to format Locally defined queries
            # Split the results and loop thru the array and apply the format to it's respective field
            if ($queryType eq "LOCAL") {
                  my @resultsFormatArray = split(/\s/, $resultsFormat);
                  my $column = 0;
                  foreach my $format (@resultsFormatArray) {
                        $bugRow[$column] = sprintf($format, $bugRow[$column]);
                        $column++;
                  }
             }

            $bugdesc[$n++] = \@bugRow;
        }
    }

    if ($returnType eq "STRING") {
        return sort @bugdesc;
    }
    return @bugdesc;
}

######################################################################
#::method _LoadCSetCache

=over 6

=item C<_LoadCSetCache(DB, Schema, entity, @ids)>

This method caches up change set information for the defect records of
immediate interest. It runs two queries whose output is used to (1)
build up a hash map from defects to their object (cc_change_set
pointer) fields and (2) a hash map from change_sets to their lists of
cc_vob_object keys. These hash tables are then used by EditAssocs
to more efficiently make changes. 

=back

=cut

######################################################################

sub _LoadCSetCache {
    my ($self, $dbName, $schema, $entityName, @idList) = @_;

    my (@data, @ccVobRec);
    my $parmSet = 0;

    # Clear existing caches

    %crChangeSets = ();
    %crDbIDs = ();
    %csetVobs = ();
    %vobDbids = ();

    # Check to see if we are doing OID query

    if (&GetParm("CQCC_QUERY_OID") eq "TRUE") {
        $parmSet = 1;
    }

    # Query 1 for defects' objects fields (cc_change_set dbids)    

    @data = $self->PerformQuery($dbName, $entityName, $schema, 
         [ "<ID> in " . join(",", @idList) ],
         [ "dbid,<ID>,cc_change_set", "%s %s %s" ], "FALSE");

    for (@data) {
        my ($defectDbID, $defectID, $objectsID) = split(" ", $_);
        if ($objectsID > 0) {
            $crChangeSets{$defectID} = $objectsID;
        }
        $crDbIDs{$defectID} = $defectDbID;
    }

    # Query 2 for changesets contents accessible by changeset dbid
    # Note: LoadCSetCache results set defined in ClassInit

    my $csets = join(",", values %crChangeSets);
    if ($csets) {
        @data = $self->PerformQuery($dbName, "cc_change_set",
            $csetSchema,
            [ "dbid in $csets" ], [ "dbid,objects", "%s %s" ], 
            "FALSE");

        my ($cset_key,@vobList);

        for (sort @data) {
            my ($key, $vob, $path) = split(" ", $_, 3);
            if (! (defined $path && $path ne "")) {  # Skip empty results
		next;
            }
            push @ccVobRec, $vob;
            my $value = $vob . " " . $path;
            if (! defined $cset_key) {
               $cset_key = $key;
            } elsif ($key ne $cset_key) {
               # Store references to temp arrays so they 
               # don't all reference the same location
               my @localVobList = @vobList;
               $csetVobs{$cset_key} = \@localVobList;
               $cset_key = $key;
               @vobList = ();
            }
            push @vobList, $value;
        }
        if (defined $cset_key) { 
            $csetVobs{$cset_key} = \@vobList;
        }
    }
    
    # Query 3 for OIDS accessible by vob dbid
    # Note: LoadCSetCache results set defined in ClassInit
   
       if ($parmSet == 1) {
       my $ccVobDbidList = join(",", @ccVobRec);
       if ($ccVobDbidList) {
            @data = $self->PerformQuery($dbName, "cc_vob_object",
            $cvobSchema,
            [ "dbid in $ccVobDbidList" ],
            [ "dbid,<NAME>,<VOBID>,<OID>", "%s,%s,%s,%s" ],
            "FALSE");

            for (@data) {
                my  ($vobDbID, $vobName, $vobid, $vobOid) = split(",", $_ , 4);
                if ($vobOid) {
                    $vobDbids{$vobDbID} = $vobOid;
                }
            }
        }
     }

    return;
}

# Internal routine used in testing
sub _RetrieveCSetCache {
    my ($self) = @_;
    return (\%crChangeSets, \%crDbIDs, \%csetVobs);
}

######################################################################
#::Tail
######################################################################

#ClassTest();
1;
